function [like,oef] = nspba(ye,Fs,yh,FsWF,pr,ext)
% Time-stamp: <2022-06-13 16:41:03 andre>
% Spectral Pedestrian Based Approach
% ye - structure response time-history data, i=1,...,N, j=1,...,n (in SI units)
% Fs - sampling frequency (Hertz)
% yh - Walking forces time-history data, i=1,...,M, (in SI units)
% FsWF - Walking forces sampling frequency (Hertz)
% pr - prior information structure array
%    f    : natural frequencies, i=1,...,m
%    dr   : modal damping ratios, i=1,...,m
%    mm   : modal masses, i=1,...,m
%    phi  : mode shapes ij, i=1,...,n; j=1,...,m (optional)
%    Sn   : PSD of noise: either j=1,...,n
% ext - extra information structure array
%    L    : measurement grid of dofs, (length >= n)
%    cdof : considered degrees of freedom (length <= n)
%    tr   : type of response, 0 displacement,1 velocity,2 acceleration
%    nr   : number of run (optional, useful for intra-subject variability)
%    cf   : cut-off frequency (optional, useful to speed up computations)
% 
% To cite:
%   Jesus, Andre (2023): Dataset for Modal testing with a pedestrian as a vibration exciter. Loughborough University. Dataset. https://doi.org/10.17028/rd.lboro.20069207
% 
% andrehjesus@gmail.com   
% Initialise variables %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
nm = length(pr.nf);  % Number of target modes
nnc = length(pr.Se); % Number of noisy channels (either 0, 1 or n)
[Ns,nd] = size(ye);  % Number of samples and DOF
nt = 3*nm+nm*nd+nnc; % Number of parameters
fr = Fs/Ns;          % Frequency resolution
% Limit frequency
if isfield(ext,'cf')
    Nf = ext.cf;             % cut-off frequency
    Nq = floor(ext.cf/fr)+1; % Nyquist index
else
    Nf = Fs/2;           % Nyquist frequency    
    Nq = floor(Ns/2)+1;  % Nyquist index
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                Upsample walking forces or structural response               %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if Fs > FsWF || Fs < FsWF
    txt = input(['WARNING! The sampling rates of walking forces ',...
                 'and structural response\n must be equal. Do you want to ',...
                 'upsample the signal with lowest rate? Y/N [Y]: '],"s");
    if isempty(txt), txt = 'Y'; end
    if txt == 'Y' || txt == 'y'
        if Fs > FsWF
            HFs = Fs; LFs = FsWF; x = yh; y='yh';
        else
            LFs = Fs; HFs = FsWF; x = ye; y='ye';
        end
        % Ratio of integers to change sampling rate
        [P,Q] = rat(HFs/LFs); tr = abs(P/Q*LFs-HFs);
        % Convert from Lower Fs to Higher Fs (Hz)
        eval([y," = resample(x,P,Q);"]);
    else
        disp('Algorithm stopped!')
        return
    end
end

%% Compute DFT and APSD of measured response %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fk = Nf*linspace(0,1,Nq);       % Compute frequency range
Ye = fft(ye);                   % DFT of data

% magnitude scaled one-sided spectrum 
P1 = 2*abs(Ye(1:Nq,:)/Ns); P1(1,:) = P1(1,:)/2;             
% Magnitude/Phase scaled one-sided spectrum (apply scaling factor)
P2 = sqrt(2/(Fs*Ns))*Ye(1:Nq,:); P2([1,end],:) = P2([1,end],:)/sqrt(2); Ye = P2.'; 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                      Computation of parameter estimates                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
hpsd = @(x) pedestrianMF(x,fr,yh,Fs,ext);
nlc = @(x) amplcons(x,nm,nd);
ObjectiveFunction = @(x)objfunc(x,Ye,Nq,Nf,nm,hpsd,ext.tr); % Objective function

% Parameters search intervals 
scf = 34/100; scf2 = 58/100; scf3 = 60/100; % variation percentage 

xmin = [(1-scf)*pr.nf(1,:)';...  % natural frequencies
        (1-scf2)*pr.dr(1,:)';... % damping ratios
        (1-scf3)*pr.mm(1,:)';... % modal masses
        -1.3*ones(nm*nd,1);...   % mode shapes
        1e-9*ones(nnc,1)];       % PSD of noise
        % 1.5e-2*ones(nm,1);...    % damping ratios

xmax = [(1+scf)*pr.nf(1,:)';...  % natural frequencies
        (1+scf2)*pr.dr(1,:)';... % damping ratios
        (1+scf3)*pr.mm(1,:)';... % modal masses
        1.2*ones(nm*nd,1);...    % mode shapes
        1e-1*ones(nnc,1)];       % PSD of noise
        % 4.2e-2*ones(nm,1);...    % damping ratios

LB = min(xmin',xmax'); UB = max(xmin',xmax'); % Lower/Upper bound       

options = optimset('Display','iter','MaxFunEvals',5000);

% gradient descent algorithm %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
disp('Computing MLE')
ini = (0.5-ext.sf/2)*(UB+LB) + ext.sf*rand(1,nt).*(UB+LB); % initial point
ini(1:3*nm) = [pr.nf';pr.dr';pr.mm'];
[like,f,ef] = fmincon(ObjectiveFunction,ini(1,:),[],[],[],[],...
                      xmin,xmax,nlc,options);
% genetic algorithms %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% $$$ pop = 100; gen = 70; time = Inf; % populations/generations
% $$$ ini = repmat(LB,pop,1) + rand(pop,nt).*repmat(UB-LB,pop,1); % initial population
% $$$ ini = repmat(LB,pop,1) + 0.6*ones(pop,nt).*repmat(UB-LB,pop,1); % initial population
% $$$ [like,f] = ga(ObjectiveFunction,nt,[],[],[],[],xmin,xmax,nlc,options);

if ef <= 0                              % not-converged
    % like = NaN(size(like));
    oef=ef; fprintf('%s','Warning! Optimisation has not converged!')
else
    oef =1;
end
return
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                             Auxiliary functions                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Objective function for  narrow band spectral model %%%%%%%%%%%%%%%%%%%%%%%%%%

function  [obj] = objfunc(x0,efft,Nq,Nf,nm,hpsd,tr)
% x0 - Modal properties vector
% efft - experimental (response) fft
% Nq - Nyquist index
% Nf - Cut-off frequency
% nm is the number of modes to identify
% tr is type of response (2 acceleration, 1 velocity, 0 displacement)
% 

% Initialise variables
obj = Inf;                % objective function initial value
[nd,Ns] = size(efft);     % number of DOF and data size
fk = Nf*linspace(0,1,Nq); % Compute frequency range

% Convert column vector to theta parameters
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fr = x0(1:nm);               % n. frequencies
dr = x0(nm+1:2*nm);          % damping ratios
mm = x0(2*nm+1:3*nm);        % modal mass
phi = x0(3*nm+1:3*nm+nd*nm); % mode shapes
phi = reshape(phi,nd,nm);    % reshape ms
Se = x0(3*nm+nd*nm+1:end);   % PSD of noise

if find([fr,dr,mm,Se]<0), return, end % truncate negative values
if length(Se) == 1, 
    Se = Se*eye(nd); 
elseif length(Se)==0
    Se = zeros(nd); 
else 
    Se = diag(Se); 
end
ipsd = hpsd(phi); % modal forces power spectral density

for k = 1:Nq                                % frequency band k    
    % Power Spectral Density of modal forces
    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    h = transf(fk(k),fr,dr,mm,tr);
    
    R = tril(ones(nm),-1); R(R==1) = ipsd(k,nm+1:end);
    epsd = diag(ipsd(k,1:nm)) + conj(R' + R);

    % Power Spectral Density of structural response
    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    E = nearestSPD(phi*diag(h)*epsd*diag(h)'*phi' + Se + ...
                   diag(eps*(nd+10)*ones(nd,1)));
    
    [Ce,pm] = chol(E,'lower'); % Cholesky decomposition of E    
    if pm, return, end         % not positive definite    
    
    cefft = Ce\(efft(:,k));    % transformation of scaled FFT

    % Likelihood function
    % %%%%%%%%%%%%%%%%%%%
    logsqrtdetpsd(k) = 2*log(prod(diag(Ce)));
    quadform(k) = cefft'*cefft;
end
% negative log-likelihood function
obj = nd*Nq*log(pi) + sum(logsqrtdetpsd) + sum(quadform); 
end

function [h] = transf(fk,fr,dr,m,q)
% Outputs a system transfer function, h, for given:
% fr   : natural frequencies, i=1,...,m
% fk   : frequency band 
% dr   : modal damping ratios, i=1,...,m
% m    : modal masses, i=1,...,m
% q    : type of response, 0 (displacement) 1 (velocity) 2 (acceleration)

% transfer function
    h = (2*pi*fk*i)^q./(4*pi^2*m'.*(fr'.^2 - fk.^2 + dr'.*fr'.*fk*2i)); 
end

function [psdmf] = pedestrianMF(phi,Fr,WF,FsWF,strinfo)
%% PSD matrix of forces - quasi-harmonic walk
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INPUT
% phi       : mode shape matrix
% Fr        : desired frequency resolution (Hz)
% WF        : Walking forces time-history data, i=1,...,M, (in SI units)
% FsWF      : sampling rate of walking forces
% strinfo - structure information array
%    L    : measurement grid of dofs, (length >= n)
%    cdof : considered degrees of freedom (length <= n)
%    nr   : number of run (optional, useful for intra-subject variability)
% OUTPUT
% psdmf : APSD and CPSD of modal forces
%
    dt=1/FsWF;
    L=strinfo.L; cdof=strinfo.cdof; nr=strinfo.nr;
    % Number of data points %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    Nx = round(1/(dt*Fr)); % number of fft points for a desired resolution
    Te = (0:Nx-1)*dt;
    F = detrend(WF); % remove static component
    F([1:(nr-1)*length(Te),nr*length(Te)+1:length(F)])=[]; % trim signal
    % mode shapes
    for i = 1:size(phi,2)
        V=phi(:,i);
        if length(V)~=length(L) %TODO: fix later
            mL=L(cdof);
        else
            mL=L;
        end
        % interpolate mode shapes %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        Ln = linspace(0,max(L),Nx); 
        % Yv(i,:) = interp1(mL,V,Ln);  % full
        % Yv(i,:) = interp1(mL,V,Ln,'linear','extrap');% extrapolation
        Yv(i,:) = interp1([0;mL;max(L)],[0;V;0],Ln);
    end
    % Modal forces %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % modulate forces according to mode shapes
    Mf = F.*Yv'; ak = fft(Mf); % modal forces + fft
    P1 = sqrt(2*dt/Nx)*ak(1:floor(Nx/2)+1,:);
    P1([1,end],:) = P1([1,end],:)/sqrt(2);  % one sided spectrum
                 
    % Auto and cross PSD
    %%%%%%%%%%%%%%%%%%%%
    psdmf = abs(P1).^2;
    for k = 1:3
        for l = k+1:3
            psdmf = [psdmf,P1(:,k).*conj(P1(:,l))];
        end
    end
end % tdm_pedestrian

function [c,ceq] = amplcons(x,nm,nd)
% Constraints of unity normalised  mode shapes max(|phi|) = 1
% nm - number of modes
% nd - number of DOFs
    msv=x(3*nm+1:(3+nd)*nm);  % mode shape vector
    phi = reshape(msv,nd,nm); % mode shapes
    mdf = max(abs(phi),[],1); % max ampl DOF

    % constraint inequality
    te = 0.05;                 % threshold error
    c = abs(msv.*repelem(1./mdf,nd)-msv) - te*ones(1,nd*nm);
    ceq = 0;
    % constraint equality
    % c = [];
    % ceq = x(3*nm+1:(3+nd)*nm).*repelem(1./mdf,nd)-x(3*nm+1:(3+nd)*nm)
end